from PIL import Image, ImageDraw, ImageFont

import textwrap
import qrcode
import math

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

from truthsayer import assets


def makeQR(data, box_size=4, border=4):
    QR = qrcode.QRCode(
        version=1,
        error_correction=qrcode.constants.ERROR_CORRECT_L,
        box_size=box_size,
        border=border,
    )
    QR.add_data(data)
    QR.make(fit=True)
    qr_code = QR.make_image(fill_color="black", back_color="white")
    return qr_code


class Renderer:
    def __init__(self, game_state, game_config, card_objects, outfile, troop_tokens=[], dead_leaders=[], quality=95, battle=False):
        self.txt_spacing_wheel = 5
        self.card_unit = 8
        self.card_radius = 10
        self.card_spacing = 10
        self.game_state = game_state
        self.game_config = game_config
        self.card_objects = card_objects
        self.battle = battle
        self.troop_edge = game_config['dimensions']['troop_edge']
        self.troop_size = game_config['dimensions']['troop']
        self.leader_size = game_config['dimensions']['leader']
        self.spice_size = game_config['dimensions']['spice']
        self.factions = game_state['meta']['factions']
        self.texts = game_state['meta']['texts']
        self.dead_leaders = dead_leaders # TODO
        self.troop_tokens = troop_tokens # TODO
        self.outfile = outfile
        self.quality = quality
        # prepare canvas
        self.prepareCanvas()
        # prepare data
        self.factions_positions = self.calculateFactionLeadersPositions()

    def prepareCanvas(self):
        filename = pkg_resources.open_binary(assets, 'map.png')
        self.canvas = Image.open(filename)
        self.canvas = self.canvas.convert('RGBA')
        self.width_canvas, self.height_canvas = self.canvas.size
        del filename
        # text layer
        self.txt = Image.new('RGBA', self.canvas.size, (255,255,255,0))
        self.d = ImageDraw.Draw(self.txt)
        filename = pkg_resources.open_binary(assets, 'FreeSans.ttf')
        self.fnt = ImageFont.truetype(filename, 15)
        del filename
        filename = pkg_resources.open_binary(assets, 'FreeSans.ttf')
        self.fnt_wheel = ImageFont.truetype(filename, 27)
        del filename
        filename = pkg_resources.open_binary(assets, 'FreeSans.ttf')
        self.fnt_card_large = ImageFont.truetype(filename, 24)
        del filename
        filename = pkg_resources.open_binary(assets, 'FreeSans.ttf')
        self.fnt_card_small = ImageFont.truetype(filename, 19)
        del filename
        filename = pkg_resources.open_binary(assets, 'FreeSans.ttf')
        self.fnt_card_tiny = ImageFont.truetype(filename, 13)
        del filename
        filename = pkg_resources.open_binary(assets, 'RobotoCondensed-Bold.ttf')
        self.fnt_troop = ImageFont.truetype(filename, 22)
        del filename

    def renderTroop(self, x, y, faction, number):
        colors = {
            'spiritual_advisor': '#274587',
            'bene_gesserit_troops': '#274587',
            'atreides_troops': '#306C3D',
            'harkonnen_troops': '#000000',
            'emperor_troops': '#ED3337',
            'sardaukar': '#ED3337',
            'spacing_guild_troops': '#E8552C',
            'fremen_troops': '#FEC64B',
            'fedaykin': '#FEC64B'
        }
        fill = colors.get(faction, 'blue')
        box = (
            x-self.troop_size/2,
            y-self.troop_size/2,
            x+self.troop_size/2,
            y+self.troop_size/2)
        self.d.ellipse(box, fill=fill, outline ='black', width=2)
        text_fill = 'white'
        if faction in ['spiritual_advisor', 'sardaukar', 'fedaykin']:
            text_fill = 'black'
            box = (
                x-self.troop_size/2+self.troop_edge,
                y-self.troop_size/2+self.troop_edge,
                x+self.troop_size/2-self.troop_edge,
                y+self.troop_size/2-self.troop_edge)
            self.d.ellipse(box, fill='white', outline ='black', width=2)
        text = str(number)
        w, h = self.fnt_troop.getsize(text)
        self.d.text((x, y+h/2-2), text, font=self.fnt_troop, fill=text_fill, anchor='ms')

    def renderTroops(self):
        areas = self.game_config['generated']['areas']['polygons']
        type_point = self.game_config['types']['areas']['point']
        for area_name, region_object in self.game_state['visual'].items():
            if area_name in type_point:
                continue
            if area_name == 'storm':
                continue
            if type(region_object) is str:
                continue
            if area_name.startswith('wheel_'):
                continue
            # print('region object')
            # print(region_object)
            for region_name, token_object in region_object.items():
                # print('token object')
                # print(token_object)
                if region_name[0] != 'R' and region_name != 'whole':
                    continue
                for token_name, token_instance in token_object.items():
                    # print('token instnace')
                    # print(token_instance)
                    x = token_instance['x']
                    y = token_instance['y']
                    token_type = token_instance['type']
                    if token_type == 'troop_token':
                        faction = token_instance['token']
                        number = token_instance['c']
                        self.renderTroop(x, y, faction, number)

    def calculateFactionLeadersPositions(self):
        positions = []
        for i in range(18):
            self.leader_r = int((self.width_canvas/2)-50)
            if i in [1, 2, 4, 5]:
                self.leader_r += 20
            reg = 2*math.pi/18
            angle = (-3*i+4)*reg+reg/2
            dx = int(self.leader_r*math.cos(angle))
            dy = int(self.leader_r*math.sin(angle))
            x = int(self.width_canvas/2)+dx-int(self.leader_size/2)
            y = int(self.height_canvas/2)+dy-int(self.leader_size/2)
            positions.append(tuple([x, y]))
        return positions

    def renderQR(self):
        if self.texts.get('qr', None) is None:
            return None
        qr_code = makeQR(self.texts['qr'])
        self.width_qr, self.height_qr = qr_code.size
        self.pos_qr_x = int(self.width_canvas-self.width_qr-20)
        self.pos_qr_y = int(self.height_canvas - self.height_qr - 40)
        self.canvas.paste(qr_code, (self.pos_qr_x, self.pos_qr_y))

        w, h = self.fnt.getsize(self.texts['promo'])
        w2, h2 = self.fnt.getsize(self.texts['qr'])
        if w2 > w:
            w = w2
        x = int(self.width_canvas-w-20)
        y = self.height_canvas - 40
        text = self.texts['promo'] + '\n' + self.texts['qr']
        self.canvas, w, h = self.renderText(text, self.fnt, 'white', x, y, anchor=None)

        text = self.texts['promo_top']
        w, h = self.fnt.getsize(text)
        x = int(self.width_canvas-w-20)
        y = self.height_canvas - self.height_qr - 40 - 20
        self.canvas, w, h = self.renderText(text, self.fnt, 'white', x, y, anchor=None)

    def renderRegionMarks(self):
        # region markings
        for i in range(18):
            color = 'black'
            r = int((self.width_canvas/2)-70)
            if i not in [12, 13]:
                r += 40
                color = 'white'
            if 6 < i < 16:
                r -= 10
            reg = 2*math.pi/18
            angle = (-i+5)*reg
            if i in [2, 4, 10, 13]:
                angle += reg/7
            else:
                angle += 6*reg/7
            dx = int(r*math.cos(angle))
            dy = int(r*math.sin(angle))
            x = int(self.width_canvas/2)+dx-4
            y = int(self.height_canvas/2)+dy
            self.d.text((x, y), 'R'+str(i+1), font=self.fnt, fill=color, anchor='ms')

    def renderFactionPositions(self):
        # faction info around the map of Arrakis
        for i in range(1, 7):
            area_name = 'player_{0}'.format(str(i))
            token_name = self.game_state['visual'].get(area_name, None)
            if token_name is None:
                continue
            filename = pkg_resources.open_binary(assets, token_name)
            token = Image.open(filename)
            token = token.convert('RGBA')
            width_token, height_token = token.size
            token = token.resize((self.leader_size, self.leader_size), Image.ANTIALIAS)
            width_token, height_token = token.size
            del filename
            x, y = self.game_config['generated']['areas']['circles'][area_name]
            half_width = int(width_token/2)
            half_height = int(height_token/2)
            if y - half_height < 0:
                y += half_height - y
            elif y + half_height > self.height_canvas:
                y -= (y + half_height) - self.height_canvas
            box_target = (x-half_width, y-half_height, x+half_width, y+half_height)
            self.canvas.paste(token, box_target, mask=token)
            # faction text info
            x_info = 0
            y_info = 0
            if i == 1:
                x_info = x + half_width + 5
                y_info = y
            if i == 4:
                x_info = x + half_width + 5
                y_info = 5
            if i == 2:
                x_info = x - half_width
                y_info = y + half_width
            if i == 6:
                x_info = x - width_token
                y_info = y + half_width
            if i == 3:
                x_info = x - half_width
                y_info = y - half_height - 20 - 20
            if i == 5:
                x_info = 5
                y_info = y + half_width
            text = self.texts['player_names'][area_name] + '\n' + self.texts['user_ids'][area_name]
            self.d.text((x_info, y_info), text, font=self.fnt, fill='white')

    def renderTleilaxuTanks(self):
        # dead leaders in the tleilaxu_tanks
        for x, y, disc_filename in self.dead_leaders:
            filename = pkg_resources.open_binary(assets, disc_filename)
            token = Image.open(filename)
            token = token.convert('RGBA')
            token = token.resize((self.leader_size, self.leader_size), Image.ANTIALIAS)
            width_token, height_token = token.size
            dx = int(self.leader_size/2)
            dy = int(self.leader_size/2)
            box_target = (x-dx, y-dy, x+dx, y+dy)
            self.canvas.paste(token, box_target, mask=token)

    def renderGameInfo(self):
        # game info
        reg = 2*math.pi/18
        angle = (-3*2+4)*reg+reg/2
        dx = int(self.leader_r*math.cos(angle))
        x_info = int(self.width_canvas/2)+dx-int(self.leader_size/2)
        y_info = 110
        game_info = 'Game {0}\nTurn {1}\n{2} phase'.format(self.texts['game_id'], str(self.texts['game_turn']), self.texts['game_phase'])
        text = game_info
        self.canvas, w, h = self.renderText(text, self.fnt, 'white', x_info, y_info, anchor=None)

    def renderLastCommands(self):
        text = '\n'.join(self.texts['commands'])
        lh = 0
        for cmd in self.texts['commands']:
            _, h = self.fnt.getsize(text)
            lh += h
        x_info = 20
        y_info = int(self.height_canvas-lh-20)
        self.canvas, _, _ = self.renderText(text, self.fnt, 'white', x_info, y_info, anchor=None)

    def render(self):
        self.placeStorm()
        self.renderRegionMarks()
        self.renderFactionPositions()
        self.renderTleilaxuTanks()
        self.renderTroops()
        self.placeSpice()
        # compose the text layer
        self.canvas = Image.alpha_composite(self.canvas, self.txt)
        self.renderBattle()
        self.renderQR()
        self.renderGameInfo()
        self.renderLastCommands()
        # remove alpha
        self.canvas = self.canvas.convert('RGB')
        self.canvas.save(self.outfile, quality=self.quality)
        del self.canvas

    def placeSpice(self):
        params = self.game_state['configs'].get('spice_token', ['czempak_spice_1', self.spice_size])
        if params[1] is None:
            params[1] = self.spice_size
        filename = params[0] + '.png'
        spice_size = params[1]
        filename = pkg_resources.open_binary(assets, filename)
        token = Image.open(filename)
        token = token.convert('RGBA')
        width_token, height_token = token.size
        token = token.resize((spice_size, spice_size), Image.ANTIALIAS)
        width_token, height_token = token.size
        del filename
        filename = pkg_resources.open_binary(assets, 'RobotoCondensed-Bold.ttf')
        fnt_troop = ImageFont.truetype(filename, 22)
        del filename
        half_width = int(width_token/2)
        half_height = int(height_token/2)
        for area_name, amount in self.game_state['visual'].items():
            if area_name.endswith('_spice'):
                x, y = self.game_config['generated']['areas']['circles'][area_name]
                box_target = (
                    int(x-width_token/2),
                    int(y-width_token/2),
                    int(x+width_token/2),
                    int(y+height_token/2))
                self.canvas.paste(token, box_target, mask=token)
                text = str(amount)
                w, h = fnt_troop.getsize(text)
                self.d.text((x+w/4, y+h/2), text, font=self.fnt_troop, fill='black', anchor='ms')

    def placeStorm(self):
        storm_object = self.game_state['visual'].get('storm', None)
        if storm_object is None:
            return
        filename = pkg_resources.open_binary(assets, storm_object['token'])
        token = Image.open(filename)
        token = token.convert('RGBA')
        width_token, height_token = token.size
        scale = storm_object['s']
        angle = storm_object['a']
        x = storm_object['x']
        y = storm_object['y']
        token = token.resize((int(scale*width_token), int(scale*height_token)), Image.ANTIALIAS)
        token = token.rotate(angle, Image.NEAREST, expand=1)
        width_token, height_token = token.size
        del filename
        x_diff = int(math.ceil(x - int(math.floor(width_token/2))))
        if x_diff < 0:
            box = (
                -x_diff,
                0,
                width_token,
                height_token
            )
            token = token.crop(box)
            width_token, height_token = token.size
            x = 0
        else:
            x -= int(math.floor(width_token/2))
        y_diff = int(math.ceil(y - int(math.floor(height_token/2))))
        if y_diff < 0:
            box = (
                0,
                -y_diff,
                width_token,
                height_token
            )
            token = token.crop(box)
            width_token, height_token = token.size
            y = 0
        else:
            y -= int(math.floor(height_token/2))
        y_diff = int(math.ceil(y - int(math.floor(height_token/2))))
        box_target = (
            int(x),
            int(y),
            int(x+width_token),
            int(y+height_token))
        self.canvas.paste(token, box_target, mask=token)

    # ref: https://code-maven.com/slides/python/rectangle-with-rounded-corners
    def round_corner(self, radius, fill):
        corner = Image.new('RGBA', (radius, radius), (0, 0, 0, 0))
        draw = ImageDraw.Draw(corner)
        draw.pieslice((0, 0, radius * 2, radius * 2), 180, 270, fill=fill)
        return corner

    # ref: https://code-maven.com/slides/python/rectangle-with-rounded-corners
    def round_rectangle(self, size, radius, fill):
        width, height = size
        rectangle = Image.new('RGBA', size, fill)
        corner = self.round_corner(radius, fill)
        rectangle.paste(corner, (0, 0))
        rectangle.paste(corner.rotate(90), (0, height - radius))  # Rotate the corner and paste it
        rectangle.paste(corner.rotate(180), (width - radius, height - radius))
        rectangle.paste(corner.rotate(270), (width - radius, 0))
        return rectangle

    def render_card(self, card_object):
        radius = self.card_radius
        fill = 'white'
        # card = self.round_rectangle(size, radius, fill)
        params = self.game_state['configs'].get('card_background', ['czempak_card_background', self.card_unit])
        filename = params[0] + '.png'
        unit = params[1]
        filename = pkg_resources.open_binary(assets, filename)
        width = 25*unit
        height = 35*unit
        size = width, height
        token = Image.open(filename)
        token = token.convert('RGBA')
        width_token, height_token = token.size
        token = token.resize((width, height), Image.ANTIALIAS)
        width_token, height_token = token.size
        del filename
        card = token
        # render card content
        text = card_object['header']
        x, y = 10, 10
        for line in textwrap.wrap(text, width=12):
            card, w, h = self.renderText(line, self.fnt_card_large, 'white', x, y, anchor=None, canvas=card)
            y += h + 9
        y += 10
        text = card_object['description']
        for line in textwrap.wrap(text, width=30):
            card, w, h = self.renderText(line, self.fnt_card_tiny, 'white', x, y, anchor=None, canvas=card)
            y += 14
        text = card_object['type']
        text_subtype = card_object.get('subtype', None)
        if text_subtype is not None:
            text += ' / ' + text_subtype
        x, y = 10, height_token - 12 - h
        card, w, h = self.renderText(text, self.fnt_card_small, 'white', x, y, anchor=None, canvas=card)
        return card, width, height

    def placeWheel(self, x, y, width, angle, username_area=None, faction_area=None, leader=None):
        filename = pkg_resources.open_binary(assets,
        'battle_wheel_numbers.png')
        token = Image.open(filename)
        token = token.convert('RGBA')
        width_token, height_token = token.size
        token = token.resize((width, width), Image.ANTIALIAS)
        token = token.rotate(angle, Image.NEAREST, expand=1)
        width_token, height_token = token.size
        del filename
        box_target = (
            int(x-width_token/2),
            int(y-height_token/2),
            int(x+width_token/2),
            int(y+height_token/2))
        self.canvas.paste(token, box_target, mask=token)
        # top of the wheel
        filename = pkg_resources.open_binary(assets,
                                             'battle_wheel_top.png')
        token = Image.open(filename)
        token = token.convert('RGBA')
        width_token, height_token = token.size
        token = token.resize((width, width), Image.ANTIALIAS)
        width_token, height_token = token.size
        del filename
        box_target = (
            int(x-width_token/2),
            int(y-height_token/2),
            int(x+width_token/2),
            int(y+height_token/2))
        self.canvas.paste(token, box_target, mask=token)
        if leader is not None:
            filename = pkg_resources.open_binary(assets, leader)
            token = Image.open(filename)
            token = token.convert('RGBA')
            token = token.resize((self.leader_size, self.leader_size), Image.ANTIALIAS)
            width_token, height_token = token.size
            dx = int(self.leader_size/2)
            dy = int(self.leader_size/2)
            half_width = int(width/4)
            box_target = (x+half_width-dx, y+half_width-dy, x+half_width+dx, y+half_width+dy)
            self.canvas.paste(token, box_target, mask=token)
        y_txt = y + self.txt_spacing_wheel
        if username_area is not None:
            text = self.game_state['visual'][username_area]
            self.canvas, w, h = self.renderText(text, self.fnt_wheel, 'black', x, y_txt, ycenter=True)
            y_txt += h
            y_txt += self.txt_spacing_wheel
        if faction_area is not None:
            text = self.game_state['visual'][faction_area]
            self.canvas, w, h = self.renderText(text, self.fnt_wheel, 'black', x, y_txt, ycenter=True)

    def renderText(self, text, font, fill, x, y, anchor='ms', ycenter=False, canvas=None):
        if canvas is None:
            canvas = self.canvas
        w, h = self.fnt_troop.getsize(text)
        txt = Image.new('RGBA', canvas.size, (255,255,255,0))
        draw = ImageDraw.Draw(txt)
        yt = y
        if ycenter:
            yt = y+h/2-2
        draw.text((x, yt), text, font=font, fill=fill, anchor=anchor)
        canvas = Image.alpha_composite(canvas, txt)
        return canvas, w, h

    def renderBattle(self):
        if not self.battle:
            return None
        TINT_COLOR = (0, 0, 0)  # Black
        TRANSPARENCY = 0.65  # Degree of transparency, 0-100%
        OPACITY = int(255 * TRANSPARENCY)
        overlay = Image.new('RGBA', self.canvas.size, TINT_COLOR+(0,))
        draw = ImageDraw.Draw(overlay)  # Create a context for drawing things on it.
        draw.rectangle(
            (
                (0, 0),
                (self.width_canvas, self.height_canvas)
            ),
            fill=TINT_COLOR+(OPACITY,))
        self.canvas = Image.alpha_composite(self.canvas, overlay)
        # place left wheel
        xthird = int(self.width_canvas/3)
        ythird = int(self.height_canvas/3)
        x = int(xthird-xthird/5)
        y = int(ythird-ythird/3)
        # attacker cards
        cards = self.game_state['areas'].get('wheel_attacker_cards', None)
        width = xthird
        if cards is not None:
            for j, key in enumerate(cards):
                card_object = self.card_objects[key]
                card, w, h = self.render_card(card_object)
                half_width = int(w/2)
                half_height = int(h/2)
                xoff = j*(w+self.card_spacing)-int(len(cards)*(w+self.card_spacing)/2)
                box_target = (x+xoff, y-half_height, x+xoff+w, y+half_height)
                self.canvas.paste(card, box_target, mask=card)
            y += width + self.card_spacing
        angle = self.game_state['visual']['wheel_attacker_value']
        leader = self.game_state['visual'].get('wheel_attacker_leader', None)
        username_area = 'wheel_attacker_player_name'
        faction_area = 'wheel_attacker_player_faction'
        self.placeWheel(x, y, width, angle, username_area=username_area, faction_area=faction_area, leader=leader)
        # place right wheel
        x = self.width_canvas-int(xthird-xthird/5)
        y = self.height_canvas-int(ythird-ythird/3)
        cards = self.game_state['areas'].get('wheel_defender_cards', None)
        if cards is not None:
            for j, key in enumerate(cards):
                card_object = self.card_objects[key]
                card, w, h = self.render_card(card_object)
                half_width = int(w/2)
                half_height = int(h/2)
                xoff = j*(w+self.card_spacing)-int(len(cards)*(w+self.card_spacing)/2)
                box_target = (x+xoff, y-half_height, x+xoff+w, y+half_height)
                self.canvas.paste(card, box_target, mask=card)
            y -= width + self.card_spacing
        angle = self.game_state['visual']['wheel_defender_value']
        leader = self.game_state['visual'].get('wheel_defender_leader', None)
        username_area = 'wheel_defender_player_name'
        faction_area = 'wheel_defender_player_faction'
        self.placeWheel(x, y, width, angle, username_area=username_area, faction_area=faction_area, leader=leader)


def generateNeighborhood(centers, locations, neighbors, regions, outfile, quality=95, skip=[]):
    dl = 10
    filename = pkg_resources.open_binary(assets, 'map.png')
    canvas = Image.open(filename)
    canvas = canvas.convert('RGBA')
    width_canvas, height_canvas = canvas.size
    del filename
    filename = pkg_resources.open_binary(assets, 'FreeSans.ttf')
    fnt = ImageFont.truetype(filename, 25)
    del filename
    # drawing layer
    layer = Image.new('RGBA', canvas.size, (255, 255, 255, 0))
    draw = ImageDraw.Draw(layer)
    # region markings
    for i in range(18):
        r = int((width_canvas/2)-25)
        reg = 2*math.pi/18
        angle = (-i+5)*reg+reg/2
        dx = int(r*math.cos(angle))
        dy = int(r*math.sin(angle))
        x = int(width_canvas/2)+dx
        y = int(height_canvas/2)+dy
        draw.text((x, y), 'R'+str(i+1), font=fnt, fill='white', anchor='ms')
    # mark centers
    for location in locations:
        if location in skip:
            continue
        x, y = centers[location]
        x = int(x)
        y = int(y)
        draw.ellipse((x-dl, y-dl, x+dl, y+dl), fill='blue', outline='blue')
        if location in regions.keys():
            regs = regions[location]
            txt = ', '.join(regs)
            if location == 'polar_sink':
                txt = 'neutral'
            draw.text((x, y+3*dl), txt, font=fnt, fill='green', anchor='ms')
    # mark neighbors
    for (loc1, loc2) in neighbors:
        if loc1 in skip or loc2 in skip:
            continue
        x1, y1 = centers[loc1]
        x1 = int(x1)
        y1 = int(y1)
        x2, y2 = centers[loc2]
        x2 = int(x2)
        y2 = int(y2)
        draw.line((x1, y1, x2, y2), fill='blue', width=2)
    # compose the text layer
    canvas = Image.alpha_composite(canvas, layer)
    # remove alpha
    canvas = canvas.convert('RGB')
    canvas.save(outfile, quality=quality)
    del canvas



import math
import random
import json

from shapely.geometry import Polygon
from shapely.geometry.point import Point

from simpleai.search.local import genetic

from brackette.memento import OriginatorJSON, Caretaker

from truthsayer.opti import TokenPlacementProblem
from truthsayer.opti import MultiTokenPlacementProblem
from truthsayer.renderer import Renderer


try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

from truthsayer import assets
from truthsayer.assets import json as json_files


class CardsManager:
    def __init__(self):
        self.deck_treachery = json.loads(pkg_resources.read_text(json_files, 'treachery_deck.json'))['treachery_deck']
        self.headers_deck_treachery = []
        for card in self.deck_treachery:
            self.headers_deck_treachery.append(card['card'])
        self.deck_alliance = json.loads(pkg_resources.read_text(json_files, 'alliance_deck.json'))['alliance_deck']
        self.headers_deck_alliance = []
        for card in self.deck_alliance:
            self.headers_deck_alliance.append(card['card'])
        self.deck_spice = json.loads(pkg_resources.read_text(json_files, 'spice_deck.json'))['spice_deck']
        self.headers_deck_spice = []
        for card in self.deck_spice:
            self.headers_deck_spice.append(card['card'])
        self.deck_generator = json.loads(pkg_resources.read_text(json_files, 'generated_decks.json'))
        self.generateStormDeck()
        self.generateTraitorDeck()
        self.generateCardIndex()

    def generateStormDeck(self):
        cards = []
        headers = []
        template = self.deck_generator['storm_deck']['template']
        values = self.deck_generator['storm_deck']['values']
        for entry in values:
            card = {
                'card': template['card'].format(*entry),
                'sectors': template['sectors'].format(*entry),
                'header': template['header'].format(*entry),
                'description': template['description'].format(*entry)
            }
            cards.append(card)
            headers.append(values[0])
        self.deck_storm = cards
        self.headers_deck_storm = headers

    def generateTraitorDeck(self):
        cards = []
        headers = []
        template = self.deck_generator['traitor_deck']['template']
        values = self.deck_generator['traitor_deck']['values']
        for entry in values:
            card = {
                'card': template['card'].format(*entry),
                'header': template['header'].format(*entry),
                'faction': template['faction'].format(*entry),
                'strength': template['strength'].format(*entry),
                'description': template['description'].format(*entry)
            }
            cards.append(card)
            headers.append(values[0])
        self.deck_traitor = cards
        self.headers_deck_traitor = headers

    def generateCardIndex(self):
        self.card_objects = {}
        all_cards = self.deck_treachery
        all_cards += self.deck_alliance
        all_cards += self.deck_spice
        all_cards += self.deck_storm
        all_cards += self.deck_traitor
        for card in all_cards:
            self.card_objects[card['card']] = card

class ConfigManager:
    def __init__(self):
        json_file = pkg_resources.read_text(assets, 'game_config.json')
        self.game_config = json.loads(json_file)
        all_areas = list(self.game_config['generated']['areas']['circles'].keys())
        all_areas += list(self.game_config['generated']['areas']['polygons'].keys())
        self.all_areas = list(set(all_areas))

    def isLeader(self, name):
        return name in self.game_config['types']['tokens']['leaders']

    def isTroop(self, name):
        return name in self.game_config['types']['tokens']['troop_tokens']

    def getRadius(self, element):
        diameter = None
        if self.isLeader(element):
            diameter = self.game_config['dimensions']['leader']
        elif self.isTroop(element):
            diameter = self.game_config['dimensions']['troop']
        radius = diameter/2
        return radius

    def getAreas(self):
        return self.all_areas

    def isAreaPoint(self, area_name):
        return area_name in self.game_config['types']['areas']['point']

    def isAreaPlayerPosition(self, area_name):
        return area_name in self.game_config['types']['areas']['players']

    def isAreaSpice(self, area_name):
        return area_name in self.game_config['types']['areas']['spice']

    def getPolygonArea(self, area_name):
        return self.game_config['generated']['areas']['polygons'][area_name]

    def getCenter(self):
         cx = self.game_config['generated']['map_center']['x']
         cy = self.game_config['generated']['map_center']['y']
         cr = self.game_config['generated']['map_center']['r']+7
         return cx, cy, cr

    def getFile(self, token_name):
        return self.game_config['files'][token_name]

    def getFactionSymbol(self, faction_name):
        return self.game_config['faction_symbols'][faction_name]

    def getFactionName(self, faction_key):
        return self.game_config['faction_names'][faction_key]

class RenderingProcessor:
    def __init__(self):
        self.manager = ConfigManager()
        self.game_config = self.manager.game_config

    def prepareInstance(self, game_state, area_name, region_name):
        polygons_maximize_overlap = Polygon(self.manager.getPolygonArea(area_name))
        # print(polygons_maximize_overlap.svg())
        # print(polygons_maximize_overlap.area)
        if region_name != 'whole':
            # print('making region')
            polygons_region = Polygon(self.manager.getPolygonArea(region_name))
            # print(polygons_region.svg())
            # print(polygons_region.area)
            polygons_maximize_overlap = polygons_maximize_overlap.intersection(polygons_region)
            # print(polygons_maximize_overlap.svg())
            # print(polygons_maximize_overlap.area)
        avoid_overlap_areas = []
        if area_name not in game_state['visual'].keys():
            game_state['visual'][area_name] = {}
        if region_name not in game_state['visual'][area_name].keys():
            game_state['visual'][area_name][region_name] = {}
        for token_name in game_state['visual'][area_name][region_name].keys():
            element_object = game_state['visual'][area_name][region_name][token_name]
            x = element_object['x']
            y = element_object['y']
            radius = self.manager.getRadius(token_name)
            polygon_token = Point(x, y).buffer(radius)
            avoid_overlap_areas.append(polygon_token)
        return polygons_maximize_overlap, avoid_overlap_areas


    def placeSingleToken(self, game_state, area_name, region_name, element_name, tolerance=0.01, amount=0):
        target_radius = self.manager.getRadius(element_name)
        polygons_maximize_overlap, avoid_overlap_areas = self.prepareInstance(game_state, area_name, region_name)
        problem = TokenPlacementProblem(polygons_maximize_overlap, avoid_overlap_areas, target_radius, tolerance=0.01)
        result = genetic(problem, population_size=75, mutation_chance=0.15, iterations_limit=120)
        x, y = result.state
        token_type = None
        if self.manager.isLeader(element_name):
            token_type = 'leader'
        if self.manager.isTroop(element_name):
            token_type = 'troop_token'
        game_state['visual'][area_name][region_name][element_name] = {
            'token': element_name,
            'type': token_type,
            'x': x,
            'y': y,
            'c': amount
        }
        return result.state


    def placeMultipleTokens(self, game_state, area_name, region_name, names, amounts):
        target_radii = [self.manager.getRadius(name) for name in names]
        polygons_maximize_overlap, avoid_overlap_areas = self.prepareInstance(game_state, area_name, region_name)
        problem = MultiTokenPlacementProblem(polygons_maximize_overlap, avoid_overlap_areas, target_radii, tolerance=0.01)
        result = genetic(problem, population_size=75, mutation_chance=0.2, iterations_limit=100)
        for i, (name, amount) in enumerate(zip(names, amounts)):
            x = result.state[i*2]
            y = result.state[i*2+1]
            token_type = None
            # print('multiplace')
            # print(name)
            if self.manager.isLeader(name):
                token_type = 'leader_token'
            elif self.manager.isTroop(name):
                token_type = 'troop_token'
            game_state['visual'][area_name][region_name][name] = {
                'token': name,
                'type': token_type,
                'x': x,
                'y': y,
                'c': amount
            }
        return result.state

    def calculateStormPosition(self, position):
        cx, cy, cr = self.manager.getCenter()
        angle = (position-6.5)*360/18
        angle_rot = angle - 90
        angle_rad = 2*math.pi*(-angle)/360
        x = cx+math.cos(angle_rad)*cr
        y = cy+math.sin(angle_rad)*cr
        storm_object = {
            'token': self.manager.getFile('storm'),
            'x': x,
            'y': y,
            's': 0.5,
            'a': angle_rot,
        }
        return storm_object

    def calculateWheel(self, value):
        return 360*(value+1.25)/21

    def process(self, game_state):
        # find objects which should be rendered but have no coordinates
        to_place = {}
        for area in self.manager.getAreas():
            if area in game_state['areas'].keys():
                if area not in game_state['visual'].keys():
                    game_state['visual'][area] = {}
                if self.manager.isAreaPoint(area):
                    if self.manager.isAreaSpice(area):
                        value = game_state['areas'][area]
                        game_state['visual'][area] = value
                    continue
                for region in game_state['areas'][area].keys():
                    if type(game_state['areas'][area][region]) is not dict:
                        continue
                    for token in game_state['areas'][area][region].keys():
                        if area in game_state['visual'].keys():
                            if region in game_state['visual'][area].keys():
                                if token in game_state['visual'][area][region].keys():
                                    old_count = game_state['visual'][area][region][token]['c']
                                    new_count = game_state['areas'][area][region][token]
                                    if new_count == 0:
                                        del game_state['visual'][area][region][token]
                                    elif old_count != new_count:
                                        game_state['visual'][area][region][token]['c'] = new_count
                                    continue
                        if area not in to_place:
                            to_place[area] = {}
                        if region not in to_place[area].keys():
                            to_place[area][region] = {}
                        to_place[area][region][token] = game_state['areas'][area][region][token]
        if 'storm' in game_state['areas'].keys():
            position = game_state['areas']['storm']
            storm_object = self.calculateStormPosition(position)
            game_state['visual']['storm'] = storm_object
        for area, faction in game_state['meta']['factions'].items():
            faction_symbol = self.manager.getFactionSymbol(faction)
            file = self.manager.getFile(faction_symbol)
            # print(faction, faction_symbol, file)
            game_state['visual'][area] = file
        wheel_values = ['wheel_attacker_value', 'wheel_defender_value']
        for wheel in wheel_values:
            if wheel in game_state['areas'].keys():
                value = game_state['areas'][wheel]
                game_state['visual'][wheel] = self.calculateWheel(value)
        wheel_leaders = ['wheel_attacker_leader', 'wheel_defender_leader']
        for wheel in wheel_leaders:
            if wheel in game_state['areas'].keys():
                leader = game_state['areas'][wheel]
                file = self.manager.getFile(leader)
                game_state['visual'][wheel] = file
        wheel_players = ['wheel_attacker_player', 'wheel_defender_player']
        wheel_players_refs = {}
        for wheel in wheel_players:
            if wheel in game_state['areas'].keys():
                player_key = game_state['areas'][wheel]
                faction_key = game_state['meta']['factions'][player_key]
                faction_name = self.manager.getFactionName(faction_key)
                player_name = game_state['meta']['usernames'][player_key]
                game_state['visual'][wheel + '_faction'] = faction_name
                game_state['visual'][wheel + '_name'] = player_name
                wheel_players_refs[wheel + '_faction'] = wheel
                wheel_players_refs[wheel + '_name'] = wheel
        # optimize positions of tokens that need placement
        for area_name in to_place.keys():
            for region_name in to_place[area_name].keys():
                elements = to_place[area_name][region_name]
                # print(elements)
                # print(len(elements.keys()))
                if len(elements.keys()) > 1:
                    amounts = []
                    names = []
                    for element_name in elements.keys():
                        element_amount = elements[element_name]
                        amounts.append(element_amount)
                        names.append(element_name)
                    self.placeMultipleTokens(game_state, area_name, region_name, names, amounts)
                    continue
                for element_name in elements.keys():
                    element_amount = elements[element_name]
                    self.placeSingleToken(game_state, area_name, region_name, element_name, amount=element_amount)
                continue
        # find objects which have coordinates but should be removed
        to_remove = []
        to_remove_points = []
        for area in game_state['visual'].keys():
            if area in ['wheel_attacker_cards', 'wheel_defender_cards']:
                continue
            if self.manager.isAreaPoint(area):
                if self.manager.isAreaPlayerPosition(area):
                    continue
                if area not in game_state['areas'].keys():
                    to_remove_points.append(area)
                continue
            if area == 'storm':
                if 'storm' not in game_state['areas'].keys():
                    to_remove_points.append(area)
                continue
            if area in wheel_values + wheel_leaders:
                if area not in game_state['areas'].keys():
                    to_remove_points.append(area)
                continue
            if area in wheel_players_refs.keys():
                wheel = wheel_players_refs[area]
                if wheel not in game_state['areas'].keys():
                    to_remove_points.append(area)
                continue
            for token_object in game_state['visual'][area]:
                if type(token_object) is str:
                    continue
                region = token_object['region']
                token = token_object['token']
                removal = None
                if area not in game_state['areas'].keys():
                    removal = area, region, token
                elif region not in game_state['areas'][area].keys():
                    removal = area, region, token
                elif token not in game_state['areas'][area][region].keys():
                    removal = area, region, token
                if removal is not None:
                    to_remove.append(removal)
        # remove gathered
        for area in to_remove_points:
            del game_state['visual'][area]
        for area, region, token in to_remove:
            del game_state['visual'][area][region][token]
        # as a last thing, generate texts
        for player_key, faction_key in game_state['meta']['factions'].items():
            player_name = game_state['meta']['usernames'][player_key]
            user_id = game_state['meta']['user_ids'][player_key]
            faction_name = self.manager.getFactionName(faction_key)
            text = '{0}\n{1}'.format(player_name, faction_name)
            if 'texts' not in game_state['meta'].keys():
                game_state['meta']['texts'] = {}
            if 'player_names' not in game_state['meta']['texts'].keys():
                game_state['meta']['texts']['player_names'] = {}
            if 'user_ids' not in game_state['meta']['texts'].keys():
                game_state['meta']['texts']['user_ids'] = {}
            game_state['meta']['texts']['player_names'][player_key] = player_name
            game_state['meta']['texts']['user_ids'][player_key] = user_id
        return game_state


class OriginatorTruthsayer(OriginatorJSON):
    def __init__(self, game_state={}, meta={}):
        self.processor = RenderingProcessor()
        self.cards_manager = CardsManager()
        if game_state == {}:
            game_state = self.initiate(meta)
        super().__init__(game_state)

    def validateMapLocation(self, target_area, target_region):
        if target_area not in self.processor.manager.getAreas():
            raise ValueError('Invalid area')
        if target_region[0] != 'R' and target_region != 'whole':
            raise ValueError('Invalid region')

    def move(self, faction, source_area, source_region, target_area, target_region, N, troop_type=None):
        self.validateMapLocation(source_area, source_region)
        self.validateMapLocation(target_area, target_region)
        if troop_type is None:
            troop_type = '{0}_troops'.format(faction)
        if source_area not in self._object_state['areas'].keys():
            raise ValueError('No troops in this area')
        if source_region not in self._object_state['areas'][source_area].keys():
            raise ValueError('No troops in this region')
        if troop_type not in self._object_state['areas'][source_area][source_region].keys():
            raise ValueError('No troops in this region')
        if self._object_state['areas'][source_area][source_region][troop_type] < N:
            raise ValueError('Insufficient amount of troops to perform this action')
        if target_area not in self._object_state['areas'].keys():
            self._object_state['areas'][target_area] = {}
        if target_region not in self._object_state['areas'][target_area].keys():
            self._object_state['areas'][target_area][target_region] = {}
        if troop_type not in self._object_state['areas'][target_area][target_region].keys():
            self._object_state['areas'][target_area][target_region][troop_type] = 0
        self._object_state['areas'][source_area][source_region][troop_type] -= N
        self._object_state['areas'][target_area][target_region][troop_type] += N
        cmd = '/{0} {1} {2} {3} {4} {5} {6}'.format('move', faction, source_area, source_region, target_area, target_region, str(N))
        self.appendCMD(cmd)

    def ship(self, faction, target_area, target_region, N, troop_type=None):
        self.validateMapLocation(target_area, target_region)
        if faction not in self._object_state['hidden']['reserves'].keys():
            raise ValueError('Incorrect faction')
        if troop_type is None:
            troop_type = '{0}_troops'.format(faction)
        if troop_type not in self._object_state['hidden']['reserves'][faction].keys():
            raise ValueError('Invalid troop type')
        if self._object_state['hidden']['reserves'][faction][troop_type] < N:
            raise ValueError('Not enough forces in reserves')
        if target_area not in self._object_state['areas'].keys():
            self._object_state['areas'][target_area] = {}
        if target_region not in self._object_state['areas'][target_area].keys():
            self._object_state['areas'][target_area][target_region] = {}
        if troop_type not in self._object_state['areas'][target_area][target_region].keys():
            self._object_state['areas'][target_area][target_region][troop_type] = 0
        self._object_state['areas'][target_area][target_region][troop_type] += N
        self._object_state['hidden']['reserves'][faction][troop_type] -= N
        cmd = '/{0} {1} {2} {3} {4}'.format('ship', faction, target_area, target_region, str(N))
        self.appendCMD(cmd)

    def change(self, target_area, target_region, N):
        pass

    # payee is the one receiving
    def pay(self, spice, payor, payee):
        if self._object_state['hidden']['spice'][payor] < spice:
            raise ValueError('Insufficient funds')
        self._object_state['hidden']['spice'][payor] -= spice
        if payee != 'spice_bank':
            self._object_state['hidden']['spice'][payee] += spice
        cmd = '/{0} {1} {2} {3}'.format('pay', spice, payor, payee)
        self.appendCMD(cmd)

    def spiceblow(self, target_area, N):
        target_area_spice = target_area + '_spice'
        if not self.processor.manager.isAreaSpice(target_area_spice):
            raise ValueError('Invalid target area')
        if target_area not in self._object_state['areas'].keys():
            self._object_state['areas'][target_area_spice] = 0
        self._object_state['areas'][target_area_spice] += N
        cmd = '/{0} {1} {2}'.format('spiceblow', target_area, str(N))
        self.appendCMD(cmd)

    def harvest(self, faction, target_area, N):
        target_area_spice = target_area + '_spice'
        if not self.processor.manager.isAreaSpice(target_area_spice):
            raise ValueError('Invalid target area')
        if target_area not in self._object_state['areas'].keys():
            self._object_state['areas'][target_area_spice] = 0
        if self._object_state['areas'][target_area_spice] < N:
            raise ValueError('Insufficient spice supply')
        self._object_state['areas'][target_area_spice] -= N
        self._object_state['hidden']['spice'][faction] += N
        cmd = '/{0} {1} {2} {3}'.format('harvest', faction, target_area_spice, str(N))
        self.appendCMD(cmd)

    def kill(self, leader):
        if not self.processor.manager.isLeader(leader):
            raise ValueError('Invalid leader')
        self._object_state['areas']['tleilaxu_tanks'][leader] = 1
        cmd = '/{0} {1}'.format('kill', leader)
        self.appendCMD(cmd)

    def revive(self, leader_or_unit, N=1):
        pass

    def lead(self, player, leader):
        participants = [
            self._object_state['areas']['wheel_attacker_player'],
            self._object_state['areas']['wheel_defender_player']
        ]
        if player not in participants:
            raise ValueError('Player is not a battle participant')
        if not self.processor.manager.isLeader(leader):
            raise ValueError('Invalid leader')
        if player == participants[0]:
            self._object_state['areas']['wheel_attacker_leader'] = leader
        if player == participants[1]:
            self._object_state['areas']['wheel_defender_leader'] = leader

    def treachery(self, player, card):
        participants = [
            self._object_state['areas']['wheel_attacker_player'],
            self._object_state['areas']['wheel_defender_player']
        ]
        if player not in participants:
            raise ValueError('Player is not a battle participant')
        faction = self._object_state['meta']['factions'][player]
        if card not in self._object_state['hidden']['cards'][faction]:
            raise ValueError('Player does not have that card in the hand')
        if card in self._object_state['hidden']['cards'][faction]: self._object_state['hidden']['cards'][faction].remove(card)
        if player == participants[0]:
            self._object_state['areas']['wheel_attacker_cards'] += [card]
        if player == participants[1]:
            self._object_state['areas']['wheel_defender_cards'] += [card]

    def discard(self, player, card):
        participants = [
            self._object_state['areas']['wheel_attacker_player'],
            self._object_state['areas']['wheel_defender_player']
        ]
        if player not in participants:
            raise ValueError('Player is not a battle participant')
        faction = self._object_state['meta']['factions'][player]
        key = 'wheel_attacker_cards'
        if player == participants[1]:
            key = 'wheel_defender_cards'
        if card not in self._object_state['areas'][key]:
            raise ValueError('This card is not part of the battle plan')
        self._object_state['areas'][key].remove(card)
        self._object_state['hidden']['discarded'].append(card)
        cmd = '/{0} {1} {2}'.format('discard', player, card)
        self.appendCMD(cmd)

    def takeback(self, player):
        participants = [
            self._object_state['areas']['wheel_attacker_player'],
            self._object_state['areas']['wheel_defender_player']
        ]
        if player not in participants:
            raise ValueError('Player is not a battle participant')
        faction = self._object_state['meta']['factions'][player]
        key = 'wheel_attacker_cards'
        if player == participants[1]:
            key = 'wheel_defender_cards'
        for card in self._object_state['areas'][key]:
            self._object_state['hidden']['cards'][faction].append(card)
        self._object_state['areas'][key] = []
        cmd = '/{0} {1}'.format('takeback', player)
        self.appendCMD(cmd)

    def battle(self, aggressor_player, defender_player):
        self._object_state['areas']['wheel_attacker_player'] = aggressor_player
        self._object_state['areas']['wheel_defender_player'] = defender_player
        self._object_state['areas']['wheel_attacker_cards'] = []
        self._object_state['areas']['wheel_defender_cards'] = []
        self._object_state['areas']['wheel_attacker_value'] = 0
        self._object_state['areas']['wheel_defender_value'] = 0
        self._object_state['areas']['wheel_attacker_leader'] = None
        self._object_state['areas']['wheel_defender_leader'] = None
        cmd = '/{0} {1} {2}'.format('battle', aggressor_player, defender_player)
        self.appendCMD(cmd)

    def deployment(self, player, N):
        participants = [
            self._object_state['areas']['wheel_attacker_player'],
            self._object_state['areas']['wheel_defender_player']
        ]
        if player not in participants:
            raise ValueError('Player is not a battle participant')
        faction = self._object_state['meta']['factions'][player]
        key = 'wheel_attacker_value'
        if player == participants[1]:
            key = 'wheel_defender_value'
        self._object_state['areas'][key] = N

    def storm(self, region):
        if region[0] != 'R':
            raise ValueError('Invalid region')
        position_str = region[1:]
        try:
            position = int(position_str)
        except:
            raise ValueError('Invalid region')
        self._object_state['areas']['storm'] = position
        cmd = '/{0} {1}'.format('storm', region)
        self.appendCMD(cmd)

    def peak(self, player, deck):
        if deck not in self._object_state['hidden']['decks'].keys():
            raise ValueError('Invalid deck name')
        faction = self._object_state['meta']['factions'][player]
        cmd = '/{0} {1} {2}'.format('peak', faction, deck)
        self.appendCMD(cmd)
        card_id = self._object_state['meta']['decks'][deck][0]
        card = self.cards_manager.card_objects[card_id]
        return card

    def draw(self, player, deck):
        if deck not in self._object_state['hidden']['decks'].keys():
            raise ValueError('Invalid deck name')
        faction = self._object_state['meta']['factions'][player]
        cmd = '/{0} {1} {2}'.format('draw', faction, deck)
        self.appendCMD(cmd)
        card_id = self._object_state['hidden']['decks'][deck].pop()
        self._object_state['hidden']['cards'][faction].append(card_id)
        card = self.cards_manager.card_objects[card_id]
        return card

    def hand(self, player):
        faction = self._object_state['meta']['factions'][player]
        return {
            'cards': self._object_state['hidden']['cards'][faction],
            'spice': self._object_state['hidden']['spice'][faction],
            'reserves': self._object_state['hidden']['reserves'][faction]
        }

    def join(self, seat, username, player_id, faction='random'):
        if len(self._object_state['meta']['usernames'].keys()) >= 6:
            raise ValueError('too many players at the table')
        if seat not in list(range(1, 7)):
            raise ValueError('seat should be an integer between 1 and 6')
        key = 'player_{0}'.format(str(seat))
        self._object_state['meta']['usernames'][key] = username
        self._object_state['meta']['user_ids'][key] = player_id
        used = []
        for _, value in self._object_state['meta']['factions'].items():
            used.append(value)
        if faction in used:
            raise ValueError('this faction is already taken')
        if faction == 'random':
            available = []
            all_factions = ['atreides', 'bene_gesserit', 'emperor', 'spacing_guild', 'fremen', 'harkonnen']
            for f in all_factions:
                if f not in used:
                    available.append(f)
            faction = random.choice(available)
        self._object_state['meta']['factions'][key] = faction
        # print()
        # print(seat, username, player_id, faction, key)
        # print(self._object_state['meta']['factions'])
        # print()
        cmd = '/{0} {1} {2} {3}'.format('join', seat, username, player_id)
        if faction is not 'random':
            cmd += ' ' + faction
        self.appendCMD(cmd)

    def randomize(self, what):
        if what == 'factions':
            used = []
            for key, value in self._object_state['meta']['factions'].items():
                used.append(value)
            random.shuffle(used)
            for key, f in zip(self._object_state['meta']['factions'].keys(), used):
                self._object_state['meta']['factions'][key] = f
            # print('randomized')
            # print(self._object_state['meta']['factions'])
            self.appendCMD('/randomize factions')

    def config(self, key, param1, param2):
        # TODO validation?
        params = [param1, param2]
        self._object_state['configs'][key] = params
        args = ['configs', key]
        cmd = '/' + 'configs'
        if param1 is not None:
            cmd += ' ' + str(param1)
        if param2 is not None:
            cmd += ' ' + str(param2)
        self.appendCMD(cmd)

    def appendCMD(self, cmd):
        self._object_state['hidden']['height'] += 1
        height = self._object_state['hidden']['height']
        self._object_state['meta']['texts']['commands'].append(str(height) + ' ' + cmd)
        if len(self._object_state['meta']['texts']['commands']) > 8:
            del self._object_state['meta']['texts']['commands'][0]
        # print(self._object_state['meta']['texts'])

    def initgame(self):
        deck_treachery = list(self.cards_manager.headers_deck_treachery)
        deck_spice = list(self.cards_manager.headers_deck_spice)
        deck_storm = list(self.cards_manager.headers_deck_storm)
        deck_traitor = list(self.cards_manager.headers_deck_traitor)
        random.shuffle(deck_treachery)
        random.shuffle(deck_spice)
        random.shuffle(deck_storm)
        random.shuffle(deck_traitor)
        meta = self._object_state['meta']
        hidden = self._object_state['hidden']
        hidden['decks'] = {
            'treachery': deck_treachery,
            'spice': deck_spice,
            'storm': deck_storm,
            'traitor': deck_traitor
        }
        areas = {
            'storm': random.randint(1, 18)
        }
        for player, faction in meta['factions'].items():
            hidden['cards'][faction] = []
            if faction == 'atreides':
                hidden['reserves']['atreides'] = {
                    'atreides_troops': 10
                }
                hidden['spice']['atreides'] = 10
                areas['arrakeen'] = {
                    'R10': {
                        'atreides_troops': 10
                    }
                }
            if faction == 'harkonnen':
                hidden['reserves']['harkonnen'] = {
                    'harkonnen_troops': 10
                }
                hidden['spice']['harkonnen'] = 10
                areas['carthag'] = {
                    'R11': {
                        'harkonnen_troops': 10
                    }
                }
            if faction == 'emperor':
                hidden['reserves']['emperor'] = {
                    'emperor_troops': 15,
                    'sardaukar': 5
                }
                hidden['spice']['emperor'] = 10
            if faction == 'bene_gesserit':
                hidden['reserves']['bene_gesserit'] = {
                    'bene_gesserit_troops': 19
                }
                hidden['spice']['bene_gesserit'] = 5
                areas['polar_sink'] = {
                    'whole': {
                        'bene_gesserit_troops': 1
                    }
                }
            if faction == 'spacing_guild':
                hidden['reserves']['spacing_guild'] = {
                    'spacing_guild_troops': 15
                }
                hidden['spice']['spacing_guild'] = 5
                areas['tueks_sietch'] = {
                    'R5': {
                        'spacing_guild_troops': 5
                    }
                }
            if faction == 'fremen':
                hidden['reserves']['fremen'] = {
                    'fremen_troops': 7,
                    'fedaykin': 3
                }
                hidden['spice']['fremen'] = 3
                a = random.randint(0, 7)
                b = random.randint(0, 7-a)
                c = random.randint(0, 7-a-b)
                areas['sietch_tabr'] = {
                    'R14': {
                        'fremen_troops': 1 + a
                    }
                }
                areas['false_wall_south'] = {
                    'R4': {
                        'fremen_troops': 1 + b
                    }
                }
                areas['false_wall_west'] = {
                    'R18': {
                        'fremen_troops': 1 + c
                    }
                }
        self._object_state['hidden'] = hidden
        self._object_state['areas'] = areas
        self.appendCMD('/init')


    def initiate(self, meta):
        hidden = {
            'reserves': {},
            'spice': {},
            'battle': {
                'aggressor': None,
                'defender': None
            },
            'cards': {},
            'discarded': [],
            'height': 0,
            'decks': {}
        }
        _object_state = {
            'hidden': hidden,
            'areas': {},
            'visual': {},
            'meta': meta,
            'configs': {}
        }
        return _object_state

    def render(self, outfile, battle=False):
       self._object_state = self.processor.process(self._object_state)
       self.backup()
       renderer = Renderer(
           self._object_state,
           self.processor.game_config,
           self.cards_manager.card_objects,
           outfile,
           battle=battle)
       renderer.render()



import math
import random

from shapely.geometry.point import Point

from simpleai.search import SearchProblem

from truthsayer.poly import generate_random


def rotateAboutPoint(ox, oy, px, py, angle):
    nx = ox + math.cos(angle)*(px-ox)-math.sin(angle)*(py-oy)
    ny = oy + math.sin(angle)*(px-ox)+math.cos(angle)*(py-oy)
    return nx, ny


def shiftFromPoint(ox, oy, px, py, delta):
    if -1 < px-ox < 1:
        # slope is practically vertical in px dimensions
        return px, px + delta
    m = (py-oy)/(px-ox)
    nx = px + delta
    ny = py + m*delta + oy
    return nx, ny


def distance(ox, oy, px, py):
    return math.sqrt((ox-px)**2+(oy-py)**2)


def mutant(state, polygon):
    ox, oy = polygon.centroid.x, polygon.centroid.y
    px, py = state
    rnd = random.random()
    if rnd < 0.5:
        angle = 2*math.pi*random.random()
        return rotateAboutPoint(ox, oy, px, py, angle)
    else:
        delta = (1 if random.random() < 0.5 else -1)*random.random()*distance(ox, oy, px, py)
        return shiftFromPoint(ox, oy, px, py, delta)


class TokenPlacementProblem(SearchProblem):
    def __init__(self, polygons_maximize_overlap, polygons_avoid_overlap_areas, target_radius, tolerance=0.1, initial_state=None):
        self.polygons_maximize_overlap = polygons_maximize_overlap
        self.polygons_avoid_overlap_areas = polygons_avoid_overlap_areas
        self.target_radius = target_radius
        self.tolerance = tolerance
        if initial_state is None:
            initial_state = self.generate_random_state()
        super().__init__(initial_state=initial_state)

    def polygonize(self, state):
        x, y = state
        state_center = Point(x, y)
        return state_center.buffer(self.target_radius)

    def heuristic(self, state):
        # how far are we from the goal?
        bad = 0
        state_polygon = self.polygonize(state)
        for avoid in self.polygons_avoid_overlap_areas:
            area = state_polygon.intersection(avoid).area
            bad += area**3
        overlap = (state_polygon.intersection(self.polygons_maximize_overlap).area)
        if overlap < state_polygon.area - self.tolerance:
            centroid = self.polygons_maximize_overlap.centroid
            px, py = state
            ox, oy = centroid.x, centroid.y
            bad += ((px - ox)**2 + (py - oy)**2)**5
        else:
            bad -= overlap
        return bad

    def crossover(self, state1, state2):
        x1, y1 = state1
        x2, y2 = state2
        rnd = random.random()
        if rnd < 0.5:
            return x1, y2
        else:
            return x2, y1

    def mutate(self, state):
        return mutant(state, self.polygons_maximize_overlap)

    def generate_random_state(self):
        state_center = generate_random(1, self.polygons_maximize_overlap, centroid=True)[0]
        state = state_center.x, state_center.y
        return state

    def value(self, state):
        # how good is this state?
        return -self.heuristic(state)


class MultiTokenPlacementProblem(SearchProblem):
    def __init__(self, polygons_maximize_overlap, polygons_avoid_overlap_areas, target_radii, tolerance=0.1, initial_state=None):
        self.N = len(target_radii)
        self.polygons_maximize_overlap = polygons_maximize_overlap
        self.polygons_avoid_overlap_areas = polygons_avoid_overlap_areas
        self.target_radii = target_radii
        self.tolerance = tolerance
        if initial_state is None:
            initial_state = self.generate_random_state()
        super().__init__(initial_state=initial_state)

    def polygonize(self, state, radius):
        x, y = state
        state_center = Point(x, y)
        return state_center.buffer(radius)

    # slow, but not meant to be used for more than
    # N=2 or N=3 polygons simultaneously
    def heuristic(self, state):
        # how far are we from the goal?
        bad = 0
        for j in range(self.N):
            px, py = state[2*j], state[2*j+1]
            radius = self.target_radii[j]
            state_polygon = self.polygonize((px, py), radius)
            for avoid in self.polygons_avoid_overlap_areas:
                area = state_polygon.intersection(avoid).area
                bad += area**3
            overlap = (state_polygon.intersection(self.polygons_maximize_overlap).area)
            if overlap < state_polygon.area - self.tolerance:
                centroid = self.polygons_maximize_overlap.centroid
                # ox, oy = centroid.x, centroid.y
                # bad += ((px - ox)**2 + (py - oy)**2)**6
                bad += state_polygon.area - overlap
                # print('punish')
            #else:
            #    bad -= overlap
            for k in range(j+1, self.N):
                pkx, pky = state[2*k], state[2*k+1]
                radiusk = self.target_radii[k]
                state_polygon_k = self.polygonize((pkx, pky), radiusk)
                collision = state_polygon.intersection(state_polygon_k).area
                bad += collision
        return bad

    def crossover(self, mother, father):
        rnd = random.random()
        child = list(mother)
        # how much inherited from father?
        N = random.randint(1, self.N)
        charm = random.sample(list(range(self.N)), N)
        for j in charm:
            child[2*j] = father[2*j]
            child[2*j+1] = father[2*j+1]
        #if N == 0:
        #    N = random.randint(0, self.N)
        #    avgs = random.sample(list(range(self.N)), N)
        #    for j in avgs:
        #        child[2*j] = (father[2*j]+mother[2*j])/2
        #        child[2*j+1] = (father[2*j]+mother[2*j])/2
        return child

    def mutate(self, state):
        N = random.randint(1, self.N)
        xmen = random.sample(list(range(self.N)), N)
        mutated = list(state)
        for j in xmen:
            px, py = state[2*j], state[2*j+1]
            nx, ny = mutant((px, py), self.polygons_maximize_overlap)
            mutated[2*j] = nx
            mutated[2*j+1] = ny
        return mutated

    def generate_random_state(self):
        state_centers = generate_random(self.N, self.polygons_maximize_overlap, centroid=True)
        state = []
        for state_center in state_centers:
            state.append(state_center.x)
            state.append(state_center.y)
        return state

    def value(self, state):
        # how good is this state?
        return -self.heuristic(state)



import random
import math
import pareto

from itertools import compress

from bs4 import BeautifulSoup

from shapely.geometry import Polygon, LineString
from shapely.affinity import translate
from shapely.geometry.point import Point

from simpleai.search.local import genetic

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

from truthsayer import assets

def extract():
    coordinates_file = pkg_resources.read_text(assets, 'arrakis.html')
    soup = BeautifulSoup(coordinates_file, 'html.parser')
    areas = {
        'circles': {},
        'polygons': {}
    }
    for area in soup.find_all('area'):
        name = area.get('title')
        shape = area.get('shape')
        coords = area.get('coords').split(',')
        coords = [int(c) for c in coords]
        if shape == 'poly':
            edges = []
            for j in range(int(len(coords)/2)):
                jx = 2*j
                jy = 2*j+1
                xy = tuple([coords[jx], coords[jy]])
                edges.append(xy)
            areas['polygons'][name] = edges
        elif shape == 'circle':
            x = coords[0]
            y = coords[1]
            center = tuple([x, y])
            areas['circles'][name] = center
    return areas


def getRegionsLocations(areas, skip=[]):
    regions = []
    locations = []
    for r in areas['polygons'].keys():
        if locations not in skip:
            if r[0] == 'R':
                regions.append(r)
            else:
                locations.append(r)
    return regions, locations


def findIntersections(areas, regions, locations, threshold=400):
    locs = {}
    for region in regions:
        region_coords = areas['polygons'][region]
        region_polygon = Polygon(region_coords)
        for location in locations:
            location_coords = areas['polygons'][location]
            location_polygon = Polygon(location_coords)
            if region_polygon.intersection(location_polygon).area > threshold:
                if location not in locs.keys():
                    locs[location] = []
                locs[location].append(region)
                locs[location] = list(set(locs[location]))
    return locs


def findNeighboring(areas, regions, locations, skip=[]):
    neighbors = []
    for i in range(len(locations)):
        loc1 = locations[i]
        if loc1 not in skip:
            loc1_coords = areas['polygons'][loc1]
            loc1_polygon = Polygon(loc1_coords)
            for j in range(i+1, len(locations)):
                loc2 = locations[j]
                if loc2 not in skip:
                    loc2_coords = areas['polygons'][loc2]
                    loc2_polygon = Polygon(loc2_coords)
                    if loc1_polygon.intersects(loc2_polygon):
                        neighbors.append(tuple([loc1, loc2]))
    return neighbors


def findCenters(areas, locations, skip=[]):
    centers = {}
    for location in locations:
        if location not in skip:
            location_coords = areas['polygons'][location]
            location_polygon = Polygon(location_coords)
            centroid = location_polygon.centroid
            centers[location] = tuple([centroid.x, centroid.y])
    return centers


def findStorm(areas):
    centers, xs, ys = [], 0, 0
    for i in range(1, 7):
        key = 'player_{0}'.format(str(i))
        player = areas['circles'][key]
        x, y = player
        centers.append(player)
        xs += x
        ys += y
    cx, cy = xs/6, ys/6
    r = 0
    for x, y in centers:
        r += math.sqrt((cx-x)**2+(cy-y)**2)
    r /= 6
    return cx, cy, r


def generate_random(number, polygon, centroid=False):
    points = []
    minx, miny, maxx, maxy = polygon.bounds
    while len(points) < number:
        pnt = Point(random.uniform(minx, maxx), random.uniform(miny, maxy))
        if polygon.contains(pnt):
            points.append(pnt)
    if len(points) == 0:
        pnt = polygon.centrod
        return [pnt]
    return points


'''
def placeToken(
        areas,
        locations,
        location_regions,
        target_location,
        target_radius,
        target_region=None,
        background=None,
        avoid_leaders=[],
        avoid_tokens=[],
        avoid_spice=[],
        avoid_zones=[],
        radius_leader=90,
        radius_token=46,
        radius_spice=46,
        w=1000,
        h=1000):
    polygons_maximize_overlap = Polygon(areas['polygons'][target_location])
    if target_region is not None:
        polygons_region = Polygon(areas['polygons'][target_region])
        polygons_maximize_overlap = polygons_maximize_overlap.intersection(polygons_region)
    avoid_overlap_areas = []
    if background is not None:
        back = Polygon(background)
        difference = back.difference(polygons_maximize_overlap)
        avoid_overlap_areas.append(difference)
    for x, y in avoid_leaders:
        center = Point(x, y)
        polygon_avoid_leader = center.buffer(radius_leader)
        avoid_overlap_areas.append(polygon_avoid_leader)
    for x, y in avoid_tokens:
        center = Point(x, y)
        polygon_avoid_token = center.buffer(radius_token)
        avoid_overlap_areas.append(polygon_avoid_token)
    for x, y in avoid_spice:
        center = Point(x, y)
        polygon_avoid_spice = center.buffer(radius_spice)
        avoid_overlap_areas.append(polygon_avoid_spice)
    for coords in avoid_zones:
        zone = Polygon(coords)
        avoid_overlap_areas.append(zone)
    problem = TokenPlacementProblem(polygons_maximize_overlap, avoid_overlap_areas, target_radius, tolerance=0.01)
    result = genetic(problem, population_size=75, mutation_chance=0.15, iterations_limit=120)
    return result.state
'''



import setuptools

setuptools.setup(
    name='truthsayer',
    version='0.0.2',
    packages=['truthsayer',],
    license='MIT',
    description = 'Visualizes and manages Dune the Board Game state',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    author = 'Marek Narozniak',
    author_email = 'marek.yggdrasil@gmail.com',
    install_requires=['pillow', 'qrcode', 'beautifulsoup4==4.8.1', 'simpleai', 'mementeur'],
    url = 'https://github.com/marekyggdrasil/arrakis',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    package_data = {
        # If any package contains *.txt or *.rst files, include them:
        '': ['assets/__init__.py', 'assets/*.png', 'assets/*.ttf', 'assets/json/*.json'],
    }
)
